@@ -27,91 +27,171 @@ file_root.mkdir(parents = True, exist_ok = True)
 
 from discordbot import *
 
+def default_logger(bot, severity, contents, no_message):
+    line = ''
+    if severity == log_severity.error:
+        line = "[ERROR]: %s" % contents
+    elif severity == log_severity.warning:
+        line = "[WARNING]: %s" % contents
+    elif severity == log_severity.important:
+        line = "[IMPORTANT]: %s" % contents
+    elif severity == log_severity.info:
+        line = "[INFO]: %s" % contents
+    elif severity == log_severity.debug:
+        line = "[DEBUG]: %s" % contents
+    else:
+        line = "[UNKNOWN SEVERITY]: %s" % contents
+    
+    print(line)
+    
 class Bot:
-    def default_logger(severity, contents, no_message):
-        line = ''
-        if severity == log_severity.error:
-            line = "[ERROR]: %s" % contents
-        elif severity == log_severity.warning:
-            line = "[WARNING]: %s" % contents
-        elif severity == log_severity.important:
-            line = "[IMPORTANT]: %s" % contents
-        else:
-            line = "[INFO]: %s" % contents
-        
-        print(line)
 
-    def __init__(self, token, sequence = 0, session_id = None, bot_identifier = "python/discordbot/v1.0"):
+    def get_active_pid():
+        active_pid = None
         try:
-            if "active" in os.listdir(file_root):
-                with open(file_root / "active", "r") as f:
-                    i = int(f.read())
-                    assert not psutil.pid_exists(i)
+            with open(file_root / "active", "r") as f:
+                active_pid = int(f.read())
         except:
-            raise error.BotAlreadyRunning("The bot appears to already be running. If the bot is not running, then please wait up to 30 seconds for the previous bot to shut down. After that, make sure that %s exists and is EMPTY (you may delete any present files)." % file_root)
+            # the file does not exist or is invalid
+            pass
+        return active_pid
+
+    def __init__(self, token, sequence = 0, session_id = None, bot_identifier = "python/discordbot/v1.1", logger = default_logger):
         
-        with open(file_root / "active", "w+") as f:
-            f.write(str(os.getpid()))
-        self.created_active = True
+        '''
+            variables
+        '''
         
+        # log
+        self.logger = logger
+        self.logger_repeated_errors = 0
+        
+        # active
+        self._is_active = False
+        
+        # bot
         self.token = token
-        self.last_received_sequence = sequence
+        self._sequence = sequence
         self.session_id = session_id
         self.bot_identifier = bot_identifier
+        self._die = False
         
-        self.periodics = []
-        self.periodic_task = None
-        self.commands = {}
-        self.event_handlers = {}
+        # gateway
+        self._gateway_endpoint = None
+        self._ws_conn = None
+        self._ws_reconnect = False
+        
+        # command
+        self._commands = {}
+        self.command_prefix = "?"
+        
+        # event
+        self._event_handlers = {}
+        
+        # rate limit
         self._rate_buckets = {}
         self._rate_bucket_lookup = {}
         
-        self.gateway_endpoint = None
-        self.gateway_connection = None
+        # periodic tasks
+        self._periodics = []
+        self._stop_periodics = False
+    
+        '''
+            activity
+        '''
+    
+        # check if it's already running
+        my_pid = os.getpid()
+        active_pid = Bot.get_active_pid()
+        if active_pid != None and psutil.pid_exists(active_pid):
+            self.log("Bot not starting because another is already running (pid = %d)" % (active_pid))
+            raise error.BotAlreadyRunning("The bot is already running elsewhere on the system. If this is not true, then please remove all files in the directory %s." % file_root)
         
-        self.ready = False
-        self.stopped = False
-        self.running = True
-        self.reconnect = False
+        # write that i'm running
+        self.log("Bot initialized.", severity = log_severity.info)
+        self.log("Writing my pid (%d) to %s" % (my_pid, file_root / "active"))
+        try:
+            with open(file_root / "active", "w+") as f:
+                f.write(str(my_pid))
+        except BaseException as e:
+            self.log_error(e)
+            raise e
+        self._is_active = True
         
-        self.logger = Bot.default_logger
         
-        # register heartbeat
-        self.register_periodic(45, self._heartbeat)
+        '''
+            create default periodic tasks
+        '''
         
-        # register logout checker
-        async def logout(dt):
-            if self.stopped:
-                await self.gateway_connection.close(1001)
-                self.gateway_connection = None
-        self.register_periodic(0.1, logout)
+        # continually check if another instance has been started
+        async def check_pid(dt):
+            active_pid = Bot.get_active_pid()
+            if active_pid is None:
+                self._is_active = False
+                self.logout()
+                raise error.DiscordError("Bot shutting down because %s was corrupted or deleted." % (file_root / "active"))
+            elif active_pid != my_pid:
+                self._is_active = False
+                self.logout()
+                raise error.BotAlreadyRunning("Another bot has been started (pid = %d) which has overrun my process (pid = %d). Logging out..." % (active_pid, my_pid))
+        self.register_periodic(1, check_pid)
         
-        # register command handler
-        self.command_prefix = "?"
+        # heartbeat
+        self._heartbeat_interval = 45 # seconds
+        self._time_since_heartbeat = 0
+        async def heartbeat(dt):
+            self._time_since_heartbeat += dt
+            if self._ws_conn and self._time_since_heartbeat > self._heartbeat_interval:
+                await self._send_payload(api.payload.OutgoingHeartbeat(self._sequence))
+                self._time_since_heartbeat = 0
+        self.register_periodic(0.1, heartbeat)
+        
+        # close connection on death
+        async def check_death(dt):
+            if self._die:
+                self._stop_periodics = True
+                try:
+                    c = self._ws_conn
+                    self._ws_conn = None
+                    await c.close(1001) # not returning
+                except:
+                    pass
+        self.register_periodic(0.1, check_death)
+        
+        '''
+            events
+        '''
+        
+        # command handler
         self.on_event(event.guild_message_sent, self._command_handler)
     
     def __del__(self):
         try:
-            if self.created_active:
+            if self._is_active:
                 os.remove(file_root / "active")
+                self._is_active = False
         except:
             pass
     
     '''
         logging
     '''
-    def log(self, txt, log_only = False, severity = log_severity.info):
-        self.logger(severity, txt, log_only or not self.ready)
+    def log(self, txt, severity = log_severity.debug):
+        try:
+            self.logger(self, severity, txt, self.logger_repeated_errors)
+        except BaseException as e:
+            self.logger_repeated_errors += 1
+            self.log_error(e)
+            self.logger_repeated_errors = 0
     
-    def log_error(self, err, log_only = False):
-        log_only = log_only or isinstance(err, error.MessageSendFailure)
-        self.log("%s\n" % traceback.format_exc() + repr(err), log_only = log_only, severity = log_severity.error)
+    def log_error(self, err, severity = log_severity.error):
+        self.log("%s\n%s" % (traceback.format_exc(), repr(err)), severity = severity)
     
     '''
-        internal utils
+        internal http utils
     '''
     def _wrap_http(self, method, *args, **kwargs):
-        self.log("Sending HTTP " + method.upper() + " request w/ args: " + ", ".join(args) + " and kwargs " + str(kwargs))
+        self.log("Sending HTTP %s request w/ args: %s and kwargs: %s" % (method.upper(), ", ".join(args), str(kwargs)))
     
         response = None
         try:
@@ -120,14 +200,9 @@ class Bot:
                 "post": requests.post,
                 "put": requests.put,
                 "delete": requests.delete,
-            }[method](*args, **kwargs)
-        except requests.exceptions.RequestException as e:
-            raise error.HTTPError(
-                "HTTP %s request failed (w/ request error %s) because: %s" % (
-                method.upper(),
-                e.__class__.__name__,
-                str(e)
-            ))
+            }[method.lower()](*args, **kwargs)
+        except requests.RequestException as e:
+            raise error.HTTPError(f"HTTP {method.upper()} request failed (w/ request error {e.__class__.__name__}) because: {str(e)}") from e
         
         sc = response.status_code
         if sc in api.status_code.http_success:
@@ -136,12 +211,12 @@ class Bot:
                 try:
                     data = response.json()
                 except:
-                    raise error.UnexpectedResponse("Invalid JSON: %s" % response.text)
+                    raise error.UnexpectedResponse(f"Invalid JSON: {response.text}")
             return data, response
         elif sc in api.status_code.http_bad_request:
-            raise error.HTTPBadRequest("Sent an invalid request, got response: %s" % response.text)
+            raise error.HTTPBadRequest(f"Sent an invalid request, got response: {response.text}")
         elif sc in api.status_code.http_insufficient_permissions:
-            raise error.HTTPBadRequest("Made request with insufficient permissions, got response: %s" % response.text)
+            raise error.HTTPBadRequest(f"Made request with insufficient permissions, got response: {response.text}")
         elif sc in api.status_code.http_try_later:
             t = 5
             if sc == api.status_code.too_many_requests:
@@ -149,10 +224,10 @@ class Bot:
                     t = response.json()['retry_after']
                 except:
                     t = 10 # just assume 10
-                    self.log("Got http rate limited, but response had no retry_after parameter; %s" % response.text, severity = log_severity.warning)
+                    self.log(f"Got http rate limited, but response had no retry_after parameter. The response was: {response.text}", severity = log_severity.warning)
                     pass
             
-            self.log("HTTP will wait %.1f seconds before retrying due to global rate limit" % t)
+            self.log("HTTP will wait %.1f seconds before retrying due to unexpected rate limit: %s" % (t, response.text))
             time.sleep(t)
             return self._wrap_http(method, *args, **kwargs)
         else:
@@ -166,7 +241,7 @@ class Bot:
             
             if bucket['remaining'] <= 0 and bucket['reset_at'] > time.time():
                 t = bucket['reset_at'] - time.time()
-                self.log("HTTP will wait %.1f seconds before retrying due to bucket rate limit" % t)
+                self.log("HTTP will wait %.1f seconds before retrying due to known bucket rate limit" % t)
                 time.sleep(t)
             
             bucket['remaining'] -= 1
@@ -206,86 +281,94 @@ class Bot:
         return data, response
     
     def _get_gateway(self):
-        data, _ = self._wrap_http("get", "https://discord.com/api/v9/gateway")
-        if not 'url' in data:
-            raise error.UnexpectedResponse("response did not contain 'url' parameter: %s" % data)
+        if self._gateway_endpoint is None:
+            data, _ = self._wrap_http("get", "https://discord.com/api/v9/gateway")
+            if not 'url' in data:
+                raise error.UnexpectedResponse("response did not contain 'url' parameter: %s" % data)
+            
+            self._gateway_endpoint = data['url']
         
-        return data['url']
+        return self._gateway_endpoint
     
     '''
         internal funcs
     '''
     async def _handle_event(self, evt, event_type, *data):
-        if event_type in self.event_handlers:
-            for h in self.event_handlers[event_type]:
+        if event_type in self._event_handlers:
+            for f in self._event_handlers[event_type]:
                 try:
-                    if asyncio.iscoroutinefunction(h):
-                        await h(evt.data, *data)
-                    else:
-                        h(evt.data, *data)
+                    await util.call_func(f, evt.data, *data)
                 except BaseException as e:
-                    self.log_error(e)
-                    raise error.CallbackError(e)
+                    raise error.CallbackError("Event callback '%s' for '%s' failed." % (repr(f), event_type)) from e
     
     async def _run_periodics(self):
-        while True:
-            await asyncio.sleep(0.01)
-            now = time.time()
-            for d in self.periodics:
-                # [dt, time_last_called, callback] = d
-                dt = now - d[1]
-                if dt >= d[0]:
-                    try:
-                        if asyncio.iscoroutinefunction(d[2]):
-                            await d[2](dt)
-                        else:
-                            d[2](dt)
-                    except KeyboardInterrupt as e:
-                        self.stopped = True
-                        return
-                    except BaseException as e:
-                        self.log_error(e)
-                    now = time.time()
-                    d[1] = now
+        self.log("Periodic tasks are now running.")
+        while not self._stop_periodics:
+            for d in self._periodics:
+                [call_rate, time_last_called, f] = d
+                
+                dt = time.time() - time_last_called
+                if dt < call_rate:
+                    continue
+                
+                try:
+                    await util.call_func(f, dt)
+                except BaseException as e:
+                    p_err = error.CallbackError("Periodic function '%s' errored." % repr(f))
+                    p_err.__cause__ = e
+                    self.log_error(p_err)
+                
+                d[1] = time.time()
+            try:
+                await asyncio.sleep(0.01)
+            except BaseException as e:
+                self.log(str(e))
+        self.log("Periodic tasks have stopped.")
     
     async def _send_payload(self, pl):
         assert isinstance(pl, api.payload.OutgoingPayload), "payload must be discordbot.api.payload.OutgoingPayload"
-        await self.gateway_connection.send(pl.dumps())
+        await self._ws_conn.send(pl.dumps())
         
     async def _wait_for_valid_payload(self):
         while True:
             try:
-                return api.payload.Payload(await self.gateway_connection.recv())
+                return api.payload.Payload(await self._ws_conn.recv())
             except error.ReceivedInvalidPayload as e:
                 self.log_error(e)
     
-    async def _wait_for_payload(self, opcode):
+    async def _wait_for_payload(self, opcode, wait_for_correct_opcode = False):
         while True:
             pl = await self._wait_for_valid_payload()
             if pl.opcode == opcode:
                 return pl
             else:
-                self.log("Waiting for payload w/ opcode %d but got one with opcode %d" % (opcode, pl.opcode))
+                s = "Waiting for payload w/ opcode %d but got one with opcode %d" % (opcode, pl.opcode)
+                self.log(s)
+                if not wait_for_correct_opcode:
+                    raise error.DiscordError(s)
     
-    async def _wait_for_valid_event(self):
+    async def _wait_for_valid_event(self, wait_for_event = False):
         while True:
             try:
-                return api.payload.Event(await self._wait_for_payload(api.opcode.event))
+                return api.payload.Event(await self._wait_for_payload(api.opcode.event, wait_for_correct_opcode = wait_for_event))
             except error.InvalidEventPayload as e:
                 self.log_error(e)
     
-    async def _wait_for_event(self, event_name):
+    async def _wait_for_event(self, event_name, wait_for_event = False, wait_for_correct_event = True):
         while True:
-            evt = await self._wait_for_valid_event()
+            evt = await self._wait_for_valid_event(wait_for_event = wait_for_event)
             if evt.name == event_name:
                 return evt
             else:
-                self.log("Waiting for event w/ name %s but got one with name %s" % (event_name, evt.name))
+                s = "Waiting for event w/ name %s but got one with name %s" % (event_name, evt.name)
+                self.log(s)
+                if not wait_for_correct_event:
+                    raise error.DiscordError(s)
     
     async def _on_payload_received(self, pl):
         if pl.opcode == api.opcode.event:
             evt = api.payload.Event(pl)
-            self.last_received_sequence = max(self.last_received_sequence, evt.sequence)
+            self._sequence = max(self._sequence, evt.sequence)
             if evt.name == 'MESSAGE_CREATE':
                 evt = api.payload.MessageCreateEvent(evt)
                 await self._handle_event(evt, event.guild_message_sent, evt.message)
@@ -310,172 +393,174 @@ class Bot:
         elif pl.opcode == api.opcode.heartbeat_ack:
             pass
         elif pl.opcode == api.opcode.reconnect or pl.opcode == api.opcode.session_invalidated:
-            self.reconnect = True
+            self._ws_reconnect = True
             if pl.opcode == api.opcode.session_invalidated:
                 self.session_id = None
         else:
             self.log("Unknown payload received: %s" % pl, severity = log_severity.warning)
     
-    async def _heartbeat(self, dt):
-        if not self.gateway_connection:
-            return
-        await self._send_payload(api.payload.OutgoingHeartbeat(self.last_received_sequence))
+    async def _open_websocket(self):
+        # get gateway endpoint
+        wss_endpoint = self._get_gateway()
+        
+        # connect to it
+        self._ws_conn = await websockets.connect(wss_endpoint)
     
-    async def _run(self):
-        # run periodic tasks
-        if self.periodic_task is None:
-            self.periodic_task = asyncio.create_task(self._run_periodics())
-        
-        # connect to gateway
+    async def _connect(self):
+        # open ws
+        self.log("Opening websocket connection...")
         try:
-            self.gateway_endpoint = self._get_gateway()
+            await self._open_websocket()
         except BaseException as e:
-            self.log_error(e)
-            return 1, "Could not find gateway endpoint"
-        try:
-            self.gateway_connection = await websockets.connect(self.gateway_endpoint)
-        except BaseException as e:
-            self.log_error(e)
-            self.gateway_connection = None
-            return 2, "Could not connect to gateway"
-        self.log("found gateway: %s" % self.gateway_endpoint)
+            raise error.GatewayError("Failed to open websocket connection.") from e
         
-        # receive the hello payload
-        hello = None
-        try:
-            hello = api.payload.Hello(await self._wait_for_payload(api.opcode.hello))
-        except BaseException as e:
-            self.log_error(e)
-            await self.gateway_connection.close(1000) # will not reconnect
-            self.gateway_connection = None
-            return 3, "Failed to receive hello payload"
-        self.log("received hello")
+        # receive hello
+        self.log("Waiting for hello payload...")
+        hello = api.payload.Hello(await self._wait_for_payload(api.opcode.hello))
+        self._heartbeat_interval = hello.heartbeat_interval
+        self.log("Discord gateway says hello! Heartbeat interval is %.1f seconds." % self._heartbeat_interval)
         
-        # update timings of periodic heartbeats
-        for d in self.periodics:
-            if d[2] == self._heartbeat:
-                d[0] = hello.heartbeat_interval
-                break
+        # send initial heartbeat
+        self.log("Replying with initial heartbeat...")
+        await self._send_payload(api.payload.OutgoingHeartbeat(self._sequence))
         
-        # initial heartbeat
-        try:
-            await self._send_payload(api.payload.OutgoingHeartbeat(self.last_received_sequence))
-        except BaseException as e:
-            self.log_error(e)
-            await self.gateway_connection.close(1000) # will not reconnect
-            self.gateway_connection = None
-            return 4, "Failed to send initial heartbeat"
-        self.log("sent initial heartbeat")
-            
         # wait for ACK
-        try:
-            await self._wait_for_payload(api.opcode.heartbeat_ack)
-        except BaseException as e:
-            self.log_error(e)
-            await self.gateway_connection.close(1000) # will not reconnect
-            self.gateway_connection = None
-            return 5, "Failed to receive initial heartbeat ACK"
-        self.log("received initial heartbeat ACK")
+        self.log("Waiting for initial ACK...")
+        await self._wait_for_payload(api.opcode.heartbeat_ack)
         
-        # resume if possible, else identify
-        if self.session_id != None:
-            # resume
-            r = api.payload.Resume(
-                self.token,
-                self.session_id,
-                self.last_received_sequence
-            )
-            try:
-                await self._send_payload(r)
-            except BaseException as e:
-                self.log_error(e)
-                await self.gateway_connection.close(1000) # will not reconnect
-                self.gateway_connection = None
-                return 6, "Failed to resume"
-            self.log("resumed using: %s" % r.dumps())
-        else:
-            # identify
-            try:
-                await self._send_payload(api.payload.OutgoingIdentify(
-                    self.token,
-                    self.intents,
-                    self.bot_identifier
-                ))
-            except BaseException as e:
-                self.log_error(e)
-                await self.gateway_connection.close(1000) # will not reconnect
-                self.gateway_connection = None
-                return 7, "Failed to identify"
-            self.log("identified")
-            
-            # wait for ready event
+        # connected!
+        self.log("Websocket connection is open and alive!")
+        return True
+    
+    async def _ws_init_identify(self):
+        # identify
+        self.log("Identifying...")
+        await self._send_payload(api.payload.OutgoingIdentify(
+            self.token,
+            self.intents,
+            self.bot_identifier
+        ))
+        
+        # wait for ready event
+        self.log("Waiting for READY event...")
+        ready = api.payload.ReadyEvent(await self._wait_for_event('READY'))
+        self.log("Logged in as %s - Got session id: %s" % (ready.user, ready.session_id))
+        self.user = ready.user
+        self.session_id = ready.session_id
+        self._sequence = 0
+        
+        return ready
+    
+    async def _ws_init_resume(self):
+        self.log("Resuming from sequence %d using session id %s..." % (self._sequence, self.session_id))
+        await self._send_payload(api.payload.Resume(
+            self.token,
+            self.session_id,
+            self._sequence
+        ))
+        self.log("Now resuming.")
+        return True
+    
+    async def _attempt_full_connect(self):
+        self.log("Attempting full gateway connection (sequence = %d, sessionid = %s)." % (self._sequence, self.session_id))
+        
+        # close the previous websocket (if there is one)
+        try:
+            if self._ws_conn != None:
+                await self._ws_conn.close(1000)
+        except:
+            # it was already closed
+            pass
+        self._ws_conn = None
+        
+        # open a new connection
+        try:
+            await self._connect()
+        except BaseException as e:
+            raise error.GatewayError("Failed to connect to gateway.") from e
+        
+        # identify or resume
+        if self.session_id == None:
             ready = None
             try:
-                ready = api.payload.ReadyEvent(await self._wait_for_event('READY'))
+                ready = await self._ws_init_identify()
             except BaseException as e:
-                self.log_error(e)
-                await self.gateway_connection.close(1000) # will not reconnect
-                self.gateway_connection = None
-                return 8, "Failed to receive ready event"
-            self.log("received ready event. session_id: %s" % ready.session_id)
-            self.user = ready.user
-            self.session_id = ready.session_id
-        
-            # run init
-            self.ready = True
+                raise error.GatewayError("Failed to identify, perhaps invalid bot token?") from e
+            
+            self.log("Running ready event after identify success.")
             await self._handle_event(ready, event.ready)
-        
-        try:
-            while True:
-                pl = await self._wait_for_valid_payload()
-                if pl.opcode == 0:
-                    self.log("received event payload: %s" % pl.raw["t"])
-                else:
-                    self.log("received payload opcode: %d - %s" % (pl.opcode, api.opcode.get_name(pl.opcode)))
-                try:
-                    await self._on_payload_received(pl)
-                except error.DiscordError as e:
-                    self.log_error(e)
-                
-                if self.reconnect:
-                    self.log("Bot will reconnect to gateway.")
-                    self.reconnect = False
-                    try:
-                        await self.gateway_connection.close(1012) # will reconnect
-                    except:
-                        pass
-                    self.gateway_connection = None
-                    return await self._run()
-        except (KeyboardInterrupt, asyncio.exceptions.CancelledError) as e:
-            self.stopped = True
+        else:
             try:
-                await self.gateway_connection.close(1000) # will not reconnect
-            except:
-                pass
-            self.gateway_connection = None
-            return 0, "KeyboardInterrupt"
-        except (BaseException, asyncio.IncompleteReadError) as e:
-            if self.stopped:
-                try:
-                    await self.gateway_connection.close(1000) # will not reconnect
-                except:
-                    pass
-                return 0, "Logged out."
-            else:
-                try:
-                    await self.gateway_connection.close(1012) # will reconnect
-                except:
-                    pass
-            self.gateway_connection = None
-            
-            if isinstance(e, asyncio.IncompleteReadError):
-                self.log("Reconnecting due to an IncompleteReadError: %s" % e)
-            else:
+                await self._ws_init_resume()
+            except BaseException as e:
+                p_err = error.GatewayError("Failed to resume connection after disconnect, perhaps the sessionid was invalidated?")
+                p_err.__cause__ = e
+                self.log_error(p_err)
+                
+                # probably invalid session id, try again without the resume
+                self.log("Resume failed - retrying without resuming (some events may be dropped).")
+                self.session_id = None
+                self._sequence = 0
+                return await self._attempt_full_connect()
+        
+        self.log("Full gateway connection successful! Now ready to receive normal payloads.")
+        return True
+        
+    async def _ws_payload_loop(self):
+        # wait for a payload
+        pl = await self._wait_for_valid_payload()
+        
+        # log it
+        if pl.opcode == api.opcode.event:
+            self.log("Received event payload: %s" % pl.raw['t'])
+        else:
+            self.log("Received payload: [%d] %s" % (pl.opcode, api.opcode.get_name(pl.opcode)))
+        
+        # process it
+        try:
+            await self._on_payload_received(pl)
+        except error.DiscordError as e:
+            self.log_error(e)
+    
+    async def _run(self):
+        self.log("Bot is running...")
+        self._periodic_task = asyncio.create_task(self._run_periodics())
+        
+        while not self._die:
+            self.log("Bot has not died, will connect and begin looping over payloads.")
+            try:
+                self._ws_reconnect = False
+                await self._attempt_full_connect()
+                while True:
+                    await self._ws_payload_loop()
+                    if self._ws_reconnect or self._die:
+                        break
+            except BaseException as e:
+                if util.error_caused_by(e, websockets.ConnectionClosedError):
+                    if self._ws_conn:
+                        if self._ws_conn.close_code == 4004:
+                            p_err = error.DiscordError("Invalid bot token.")
+                            p_err.__cause__ = e
+                            self.log_error(p_err)
+                            return False
+                        elif self._ws_conn.close_code == 4009:
+                            self.session_id = None
+                        self.log("Bot will reconnect due to websocket closing with code %d." % self._ws_conn.close_code)
+                    self._ws_reconnect = True
+                
+                if self._ws_reconnect:
+                    continue
+                
+                if util.error_caused_by(e, (KeyboardInterrupt, asyncio.CancelledError)):
+                    self._die = True
+                    self.log("Bot shutting down due to KeyboardInterrupt or asyncio.CancelledError.", severity = log_severity.info)
+                    return True
+                
                 self.log_error(e)
-                self.log("Due to above error, the bot will reconnect to the gateway.", severity = log_severity.important)
-            
-            self.ready = False
-            return await self._run()
+                return False
+                
+        self.log("Bot died due to normal operations.")
+        return True
     
     def _command_handler(self, raw, message):
         if not message.content.startswith(self.command_prefix):
@@ -491,11 +576,11 @@ class Bot:
         sent_cmd = parts[0][len(self.command_prefix):]
         sent_args = parts[1:]
         
-        for cmd in self.commands:
+        for cmd in self._commands:
             if not sent_cmd in cmd.split("_,_"):
                 continue
             
-            cmd_info = self.commands[cmd]
+            cmd_info = self._commands[cmd]
             if cmd_info['required_role'] != None and not cmd_info['required_role'] in roles:
                 return
             
@@ -551,18 +636,50 @@ class Bot:
     '''
     def on_event(self, evt, callback):
         assert issubclass(evt, event.Base), "event must be a discordbot.event"
-        if not evt in self.event_handlers:
-            self.event_handlers[evt] = []
-        self.event_handlers[evt].append(callback)
+        if not evt in self._event_handlers:
+            self._event_handlers[evt] = []
+        self._event_handlers[evt].append(callback)
     
     def remove_callback(self, evt, callback):
-        assert evt in self.event_handlers and callback in self.event_handlers[evt], "event callback is not registered, you must Bot.on_event() it first"
-        self.event_handlers[evt].remove(callback)
-        if len(self.event_handlers[evt]) == 0:
-            del self.event_handlers[evt]
+        assert evt in self._event_handlers and callback in self._event_handlers[evt], "event callback is not registered, you must Bot.on_event() it first"
+        self._event_handlers[evt].remove(callback)
+        if len(self._event_handlers[evt]) == 0:
+            del self._event_handlers[evt]
     
     def register_periodic(self, dt, callback):
-        self.periodics.append([dt, time.time(), callback])
+        self._periodics.append([dt, time.time(), callback])
+    
+    def add_role(self, guild, user, role):
+        if isinstance(guild, api.structure.Guild):
+            guild = guild.id
+        if isinstance(user, api.structure.User):
+            user = user.id
+            
+        data, _ = self._rate_limited_http(
+            "put",
+            "https://discord.com/api/v9/guilds/%s/members/%s/roles/%s" % (guild, user, role),
+            headers = {
+                "Authorization": "Bot %s" % self.token,
+                "Content-Type": "application/json"
+            }
+        )
+        return data
+    
+    def remove_role(self, guild, user, role):
+        if isinstance(guild, api.structure.Guild):
+            guild = guild.id
+        if isinstance(user, api.structure.User):
+            user = user.id
+            
+        data, _ = self._rate_limited_http(
+            "delete",
+            "https://discord.com/api/v9/guilds/%s/members/%s/roles/%s" % (guild, user, role),
+            headers = {
+                "Authorization": "Bot %s" % self.token,
+                "Content-Type": "application/json"
+            }
+        )
+        return data
     
     def send_message(self, channel, content = '', embeds = [], reply_to = None, force_reply = False, reply_ping = None, allowed_mentions = None):
         if isinstance(channel, api.structure.Channel):
@@ -720,8 +837,12 @@ class Bot:
         return data
     
     def stop(self):
-        self.stopped = True
+        self._die = True
     logout = stop
+    kill = stop
+    
+    def reconnect(self):
+        self._ws_reconnect = True
     
     def register_command(self, cmd, description, handler, required_role = None, args = []):
         assert isinstance(cmd, (list, str)), "cmd may only be a string or list of strings"
@@ -729,9 +850,9 @@ class Bot:
             assert all(map(lambda x: isinstance(x, str), cmd)), "list of commands must only contain strings"
         
         cmd_key = cmd if isinstance(cmd, str) else "_,_".join(cmd)
-        if cmd_key in self.commands:
+        if cmd_key in self._commands:
             self.log("called register_command with command name that is already registered (\"%s\"). Overwriting." % cmd_key, severity = log_severity.warning)
-        self.commands[cmd_key] = {
+        self._commands[cmd_key] = {
             "description": str(description),
             "handler": handler,
             "required_role": str(required_role) if required_role != None else None,
@@ -759,7 +880,7 @@ class Bot:
                                 "@everyone" if info["required_role"] is None else util.mention_role(info["required_role"]),
                                 info["description"]
                             )
-                        } for (cmd, info) in self.commands.items() 
+                        } for (cmd, info) in self._commands.items() 
                     ]
                 }
             ], allowed_mentions = {
@@ -767,9 +888,11 @@ class Bot:
             }, reply_to = message, reply_ping = False)
         self.register_command("cmds", "Shows a list of the bot's commands", cmd_cmds)
     
-    def run(self, intents):
+    def run(self, intents = [intent.guild_messages]):
         self.intents = util.get_flags(*intents)
-        self.running = True
-        a,b = asyncio.run(self._run())
-        self.running = False
-        return a,b
\ No newline at end of file
+        success = False
+        try:
+            success = asyncio.run(self._run())
+        except KeyboardInterrupt as e:
+            success = True
+        return success
\ No newline at end of file
